userTable = [
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'},
]

userTable_missing_values = [
    {'age': 18, 'location': 'Korea'},
    {'name': 'Charse', 'gender': 'M', 'location': 'USA'},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
    {'name': 'Xiaomi', },
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
    {'location': 'Korea'},
]

userTable_has_one_column_name = [
    {'name': 'Charse'},
    {'name': 'Lyn'},
    {'name': 'Xiaomi'},
    {'name': 'Park'},
    {'name': 'Smith'},
]

buyTable = [
    {'name': 'Xiaomi', 'product': 'battery', 'amount':7},
    {'name': 'Hong', 'product': 'keyboard', 'amount':1},
    {'name': 'Lyn', 'product': 'cleaner', 'amount':5},
    {'name': 'Hong', 'product': 'monitor', 'amount':1},
    {'name': 'Hong', 'product': 'mouse', 'amount':3},
    {'name': 'Lyn', 'product': 'mouse', 'amount':1},
    {'name': 'Unknown', 'product': 'keyboard', 'amount':1},
    {'name': 'Lee', 'product': 'hardcase', 'amount':2},
    {'name': 'Lee', 'product': 'keycover', 'amount':2},
    {'name': 'Yuki', 'product': 'manual', 'amount':1},
    {'name': 'Xiaomi', 'product': 'cable', 'amount':1},
    {'name': 'anonymous', 'product': 'adopter', 'amount':2},
    {'name': 'Park', 'product': 'battery', 'amount':2},
    {'name': 'Hong', 'product': 'cleaner', 'amount':3},
    {'name': 'Smith', 'product': 'mouse', 'amount':1},
]



import csv
from ..records import get_allkeys
from .io import get_stringio, reduce_csv_input



def read_csv(file, encoding='utf-8', fields=None, **csv_kwargs):
    fp = reduce_csv_input(file, encoding=encoding)
    csv_reader = csv.reader(fp, **csv_kwargs)
    fields = fields or next(csv_reader)
    result = [dict(zip(fields, map(str, row))) for row in csv_reader]
    return result



def write_csv(records:list[dict], filename=None, fields=None, encoding='utf-8', **csv_kwargs):
    csv_kwargs.update({'lineterminator':'\n'})
    output = get_stringio()
    fields = fields or get_allkeys(records)
    writer = csv.DictWriter(output, fields, **csv_kwargs)
    writer.writeheader()
    for row in records:
        writer.writerow(row)
    if filename:
        with open(filename, 'w', encoding=encoding) as fp:
            fp.write(output.getvalue())
    else:
        return output.getvalue()



import pytest

from listorm.api.records import *
from .samples import *


test_select_cases = [
    (userTable, 'name', ['location'], None, userTable_select_name_exclude_location),
    (userTable, [], None, lambda age: age > 20, userTable_age_over_20),
    (userTable, ['name', 'gender', 'location' ], None, lambda age: age > 20, userTable_select_name_gender_location_where_age_gt_20),
]
@pytest.mark.parametrize('records, columns, excludes, where, results', test_select_cases)
def test_select(records, columns, excludes, where, results):
    lst = select(records, columns, excludes=excludes, where=where)
    assert results == lst

test_drop_column_cases = [
    (userTable, ['age'], userTable_dropped_age),
    (userTable, ['gender', 'location'], userTable_dropped_gender_location),
]
@pytest.mark.parametrize('records, columns, results', test_drop_column_cases)
def test_drop(records, columns, results):
    lst = drop(records, columns)
    assert results == lst

test_add_column_cases = [
    (userTable, {'key': 'value'}, userTable_column_added_value),
    (userTable, {'gender/age': lambda gender, age: '{}/{}'.format(gender, age)}, userTable_column_added_gender_age_concated)
]
@pytest.mark.parametrize('records, column_mapset, results', test_add_column_cases)
def test_add_column(records, column_mapset, results):
    lst = add_column(records, **column_mapset)
    assert results == lst

test_rename_cases = [
    (userTable, {'location': 'address'}, userTable_renamed_location2address),
    (userTable, {'location': 'address', 'gender': 'sex'}, userTable_renamed_location2address_gender2sex),
]
@pytest.mark.parametrize('records, renamemap, results', test_rename_cases)
def test_rename(records, renamemap, results):
    lst = rename(records, **renamemap)
    assert results == lst

test_update_cases = [
    (userTable, {'name': str.upper}, None, userTable_updated_name_upper),
    (userTable, {'name': str.upper}, lambda age: age > 20, userTable_updated_name_upper_where_age_gt_20),
    (userTable, {'gender': lambda gender, age: "{}/{}".format(gender, age) }, None, userTable_updated_gender_concat_with_age),
]
@pytest.mark.parametrize('records, updatemap, where, results', test_update_cases)
def test_update(records, updatemap, where, results):
    lst = update(records, where=where, **updatemap)
    assert results == lst


test_values_cases = [
    (userTable, ['name'], True, userTable_values_by_name),
    (userTable, ['name'], False, userTable_values_by_name_flat_flase),
    (userTable, ['name', 'gender'], True, userTable_values_by_name_and_gender),
]
@pytest.mark.parametrize('records, columns, flat_one, results', test_values_cases)
def test_values(records, columns, flat_one, results):
    lst = values(records, columns, flat_one=flat_one)
    assert results == lst

test_distinct_cases = [
    (buyTable, ['name'], True, False, buyTable_distinct_by_name_keep_first),
    (buyTable, ['name'], False, False, buyTable_distinct_by_name_keep_last),
    (buyTable, ['name'], True, True, buyTable_distinct_by_name_singles),
]
@pytest.mark.parametrize('records, columns, first, singles, results', test_distinct_cases)
def test_distinct(records, columns, first, singles, results):
    lst = distinct(records, columns, keep_first=first, singles=singles)
    assert results == lst

test_orderby_cases = [
    (userTable, ['location', '-age'], userTable_orderby_location_age_desc)
]
@pytest.mark.parametrize('records, sortkeys, results', test_orderby_cases)
def test_orderby(records, sortkeys, results):
    lst = orderby(records, sortkeys)
    assert results == lst

test_groupby_cases = [
    (userTable, ['location', 'gender'], {'gender': len, 'age': lambda vals: sum(vals)/len(vals)}, {'gender': 'count', 'age': 'avg_age'}, userTable_groupby_location_gender)
]
@pytest.mark.parametrize('records, columns, aggset, renames, results', test_groupby_cases)
def test_groupby(records, columns, aggset, renames, results):
    lst = groupby(records, columns, renames=renames, **aggset)
    assert results == lst

test_join_cases = [
    (userTable, buyTable, 'name', None, 'inner', user_buy_table_inner_joined),
    (userTable, buyTable, 'name', None, 'left', user_buy_table_left_joined),
    (userTable, buyTable, 'name', None, 'right', user_buy_table_right_joined),
    (userTable, buyTable, 'name', None, 'outer', user_buy_table_outer_joined),
]
@pytest.mark.parametrize('records, other, on, right_on, how, results', test_join_cases)
def test_join(records, other, on, right_on, how, results):
    lst = join(records, other, on, on, right_on, how)
    lst = fillmissed(lst, '')
    lst = orderby(lst, 'name', 'product')
    lst = set_number_format(lst, amount=0)
    for rec, l in zip(results, lst):
        assert rec == l


'''
Functional API for list as records
----------------------------------
'''

from itertools import tee
from collections import abc, namedtuple

from .row import *
from ..utils import reduce_args, reduce_kwargs, pluralize_params
from .helper import reduce_where
from ..exceptions import UniqueConstraintError


@reduce_args('keys')
def values(records:list[dict], keys:list, *, flat_one=True) -> list[tuple]:
    '''extract values list from records

    :param records: a list contains dict items
    :param keys: keys for extracting values
    :param flat_one: If True, when the return value is a single value, the value is returned without being included in the list, defaults to True
    :return: list contains tuple values


    .. doctest::
        
        >>> import listorm

        >>> userTable = [
        ...    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
        ...    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
        ...    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
        ...    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
        ...    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
        ...    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
        ...    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'},
        ... ]

        >>> records = listorm.values(userTable, 'name', 'gender', 'location')
        >>> for row in records: print(row)
        ('Hong', 'M', 'Korea')
        ('Charse', 'M', 'USA')
        ('Lyn', 'F', 'China')
        ('Xiaomi', 'M', 'China')
        ('Park', 'M', 'Korea')
        ('Smith', 'M', 'USA')
        ('Lee', 'F', 'Korea')

        >>> # When there is only one column, a flatten single list is returned
        >>> listorm.values(userTable, 'name')
        ['Hong', 'Charse', 'Lyn', 'Xiaomi', 'Park', 'Smith', 'Lee']

        >>> # but, flat_one = False when consistency is required
        >>> listorm.values(userTable, 'name', flat_one=False)
        [('Hong',), ('Charse',), ('Lyn',), ('Xiaomi',), ('Park',), ('Smith',), ('Lee',)]

    '''
    return [
        asvalues(row, keys, flat=flat_one) for row in records
    ]


@reduce_args('keys')
def select(records:list[dict], keys:list, *, excludes:list=None, where:callable=None) -> list[dict]:
    '''Retrieves the item of the specified item

    :param records: a list contains dict items
    :param keys: keys for selecting item
    :param excludes: keys for excluded, defaults to None
    :param where: callback for filtering records
    :return: list of selected and filtered as records


    .. doctest::

        >>> # Names of people over 20 years of name, age, location
        >>> records = listorm.select(userTable, ['name', 'age', 'location'], where=lambda age: age > 20)
        >>> for row in records: row
        {'name': 'Lyn', 'age': 28, 'location': 'China'}
        {'name': 'Park', 'age': 29, 'location': 'Korea'}

    '''
    return [
        asselect(row, keys, excludes=excludes) for row in records
        if reduce_where(row, where)
    ]


@reduce_kwargs('updatemap')
def update(records:list[dict], updatemap:dict=None, where:callable=None):
    '''update item values

    :param records: a list contains dict items
    :param updatemap: key:value|callable mapping for update values
    :param where: filter specifying what to update, defaults to None
    :return: list of updated items as records


    .. doctest::

        >>> # Increase the age of people in Korea and China by one
        >>> records = listorm.update(
        ...     userTable,
        ...     age=lambda age: age +1,
        ...     # or updatemap={'age': lambda age: age +1} 
        ...     where=lambda location: location.lower() in ['korea', 'china']
        ... )
        >>> for row in records: print(row)
        {'name': 'Hong', 'gender': 'M', 'age': 19, 'location': 'Korea'}
        {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'}
        {'name': 'Lyn', 'gender': 'F', 'age': 29, 'location': 'China'}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 16, 'location': 'China'}
        {'name': 'Park', 'gender': 'M', 'age': 30, 'location': 'Korea'}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}
        {'name': 'Lee', 'gender': 'F', 'age': 13, 'location': 'Korea'}
 
        # applying value modifing functions
        >>> records = listorm.update(userTable,
        ...     # age=float, location=str.upper
        ...     updatemap={'age': float, 'location': str.upper} 
        ... )

        >>> for row in records: row
        {'name': 'Hong', 'gender': 'M', 'age': 18.0, 'location': 'KOREA'}
        {'name': 'Charse', 'gender': 'M', 'age': 19.0, 'location': 'USA'}
        {'name': 'Lyn', 'gender': 'F', 'age': 28.0, 'location': 'CHINA'}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15.0, 'location': 'CHINA'}
        {'name': 'Park', 'gender': 'M', 'age': 29.0, 'location': 'KOREA'}
        {'name': 'Smith', 'gender': 'M', 'age': 17.0, 'location': 'USA'}
        {'name': 'Lee', 'gender': 'F', 'age': 12.0, 'location': 'KOREA'}


    .. note::

        *packed* vs *unpacked* the two way of parameterization supported

        list for columns args and dict for mapping args can be converted into variable args or kwargs parameters in this library

        The following below examples will use the unpacked parameters, or if unpacked is not available, you will see it used in packed form.


        .. code-block::

            # unpacked args vs packed to list
            listorm.select(userTable, 'name', 'gender') == listorm.select(userTable, ['name', 'gender'])
        
        .. code-block::

            # unpacked kwargs vs packed to dict
            listorm.update(userTable, name=str.upper) == listorm.update(userTable, updatemap={'name': str.upper})

    '''
    return [
        asupdate(row, updatemap=updatemap) if reduce_where(row, where) else row
        for row in records
    ]


@reduce_kwargs('keymap')
def add_column(records:list[dict], *, keymap:dict) -> list[dict]:
    '''add keys to item in records

    :param records: a list contains dict items
    :param keymap: the value or function mapping required for key to be added
    :return: extended records


    .. doctest::

        >>> # Create a new column using the other two columns
        >>> records = listorm.add_column(userTable, 
        ...     keymap={'gender/age': lambda gender, age: "{}/{}".format(gender, age)}
        ... )
        >>> for row in records: row
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'gender/age': 'M/18'}
        {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA', 'gender/age': 'M/19'}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'gender/age': 'F/28'}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'gender/age': 'M/15'}
        {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'gender/age': 'M/29'}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'gender/age': 'M/17'}
        {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'gender/age': 'F/12'}

    .. note::
        As above, if a slash or space is included in the column name, unpacked paramater is not allowed


        .. code-block::

            # due to / is not allowed to using as parameter name, this cannot be valid code
            add_column(userTable, gender/age=lambda gender, age: "{}/{}".format(gender, age))
    
    '''

    return [
        addkeys(row, keymap=keymap) for row in records
    ]


@reduce_args('keys')
def drop(records:list[dict], keys:list) -> list:
    '''delete key from the items in records

    :param records: a list contains dict items
    :param keys: keys for to delete
    :return: list what contains deleted items as records


    .. doctest::

        >>> records = listorm.drop(userTable, 'age', 'gender')
        >>> for row in records: row
        {'name': 'Hong', 'location': 'Korea'}
        {'name': 'Charse', 'location': 'USA'}
        {'name': 'Lyn', 'location': 'China'}
        {'name': 'Xiaomi', 'location': 'China'}
        {'name': 'Park', 'location': 'Korea'}
        {'name': 'Smith', 'location': 'USA'}
        {'name': 'Lee', 'location': 'Korea'}

    .. note::

        using by select and excludes parameter, can retrive same results


        .. code-block::

            listorm.select(userTable, excludes=['age', 'gender'])

    '''
    return select(records, excludes=keys)


@reduce_kwargs('renamemap')
def rename(records:list[dict], renamemap:dict) -> dict:
    '''change key of items in records as to another name

    :param records: a list contains dict items
    :param renamemap: old: new key mapping dict
    :return: list what contains renamed items as records


    .. doctest::

        >>> records = listorm.rename(userTable, gender='sex', location='country')
        >>> for row in records: row
        {'name': 'Hong', 'sex': 'M', 'age': 18, 'country': 'Korea'}
        {'name': 'Charse', 'sex': 'M', 'age': 19, 'country': 'USA'}
        {'name': 'Lyn', 'sex': 'F', 'age': 28, 'country': 'China'}
        {'name': 'Xiaomi', 'sex': 'M', 'age': 15, 'country': 'China'}
        {'name': 'Park', 'sex': 'M', 'age': 29, 'country': 'Korea'}
        {'name': 'Smith', 'sex': 'M', 'age': 17, 'country': 'USA'}
        {'name': 'Lee', 'sex': 'F', 'age': 12, 'country': 'Korea'}

        >>> # if key name has to be containing spaces
        >>> records = listorm.rename(userTable, renamemap={'age': 'age of ultron' })
        >>> for row in records: row
        {'name': 'Hong', 'gender': 'M', 'age of ultron': 18, 'location': 'Korea'}
        {'name': 'Charse', 'gender': 'M', 'age of ultron': 19, 'location': 'USA'}
        {'name': 'Lyn', 'gender': 'F', 'age of ultron': 28, 'location': 'China'}
        {'name': 'Xiaomi', 'gender': 'M', 'age of ultron': 15, 'location': 'China'}
        {'name': 'Park', 'gender': 'M', 'age of ultron': 29, 'location': 'Korea'}
        {'name': 'Smith', 'gender': 'M', 'age of ultron': 17, 'location': 'USA'}
        {'name': 'Lee', 'gender': 'F', 'age of ultron': 12, 'location': 'Korea'}

        .. note::

            In newer versions of python, the order of key occurence in dict is valid

    '''
    return [
        asrename(row, renamemap=renamemap)
        for row in records   
    ]

def get_allkeys(records:list[dict]) -> list:
    '''extract the occurance keys

    :param records: a list contains dict items
    :return: a list of key occurance


    .. doctest::

        >>> arecords = [
        ...    {'name': 'Hong', 'gender': 'M'},
        ...    {'name': 'Xiaomi', 'location': 'China'}
        ... ]

        >>> listorm.get_allkeys(arecords)
        ['name', 'gender', 'location']
    

    '''
    keyset = {
        key:None
        for row in records
        for key in row
    }
    return list(keyset)


def fillmissed(records:list[dict], value=None):
    '''fill missing key as default value for normalizing records

    :param records: a list contains dict items
    :param value: default value if item has missing key, defaults to None
    :return: a records that filled with default values


    .. doctest::

        >>> userTable_with_missing_values = [
        ...     {'age': 18, 'location': 'Korea'},
        ...     {'name': 'Charse', 'gender': 'M', 'location': 'USA'},
        ...     {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
        ...     {'name': 'Xiaomi', },
        ...     {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
        ...     {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
        ...     {'location': 'Korea'},
        ... ]

        >>> # as default, missing values are filled with None
        >>> records = listorm.fillmissed(userTable_with_missing_values)
        >>> for row in records: row
        {'age': 18, 'location': 'Korea', 'name': None, 'gender': None}
        {'name': 'Charse', 'gender': 'M', 'location': 'USA', 'age': None}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'}
        {'name': 'Xiaomi', 'age': None, 'location': None, 'gender': None}
        {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}
        {'location': 'Korea', 'age': None, 'name': None, 'gender': None}

        >>> # or filled with 'undefined'
        >>> records = listorm.fillmissed(userTable_with_missing_values, 'undefined')
        >>> for row in records: row
        {'age': 18, 'location': 'Korea', 'name': 'undefined', 'gender': 'undefined'}
        {'name': 'Charse', 'gender': 'M', 'location': 'USA', 'age': 'undefined'}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'}
        {'name': 'Xiaomi', 'age': 'undefined', 'location': 'undefined', 'gender': 'undefined'}
        {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}
        {'location': 'Korea', 'age': 'undefined', 'name': 'undefined', 'gender': 'undefined'}

        >>> # rearrange key order as same as original userTable by using select
        >>> records = listorm.select(records, listorm.get_allkeys(userTable))
        >>> for row in records: row
        {'name': 'undefined', 'gender': 'undefined', 'age': 18, 'location': 'Korea'}
        {'name': 'Charse', 'gender': 'M', 'age': 'undefined', 'location': 'USA'}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'}
        {'name': 'Xiaomi', 'gender': 'undefined', 'age': 'undefined', 'location': 'undefined'}
        {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}
        {'name': 'undefined', 'gender': 'undefined', 'age': 'undefined', 'location': 'Korea'}

    '''

    tee1, tee2 = tee(records, 2)
    fields = get_allkeys(tee1)
    defaults = dict.fromkeys(fields, value)
    return [
        asdefault(row, defaultmap=defaults) for row in tee2
    ]


def guess_type(records:list[dict], key:str):
    values = values(records, key)
    head, *_ = filter(None, values)
    return type(head)


@reduce_args('sortkeys')
def orderby(records:list[dict], sortkeys:list) -> list[dict]:
    '''sort items in records

    :param records: a list of dict items
    :param sortkeys: key for sorting
    :return: sorted records


    .. doctest::


        >>> # complex key sorting: location as ascending and by age decending
        >>> records = listorm.orderby(userTable, 'location', '-age')
        >>> for row in records: row
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'}
        {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'}
        {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'}
        {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

        >>> # ordering by callback
        >>> # order by first digit of age
        >>> records = listorm.orderby(userTable, lambda age: str(age)[-1])
        >>> for row in records: row
        {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'}
        {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'}
        {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'}


    .. note::

        If you are familiar with django it looks like it

    '''
    records = list(records)
    if not records:
        return records

    for keys in reversed(sortkeys):
        reverse = False
        if isinstance(keys, str):
            if keys.startswith('-'):
               keys = keys[1:]
               reverse = True
            records.sort(key=lambda x: x[keys], reverse=reverse)
        elif callable(keys):
            records.sort(key=lambda row: reduce_callback(row, app=keys))
    return records


@reduce_args('keys')
def distinct(records:list[dict], keys:list, *, keep_first:bool=True, singles:bool=False) -> list:
    '''remove duplicated rows by keys

    :param records: a list contains dict items
    :param keys: key, check for duplicated values
    :param keep_first: if True, in case of duplicate occurrence select the first item that appears, else last, defaults to True
    :param singles: If True, eliminate rows that have been duplicated. That is, only items that were unique are left. defaults to False, defaults to False
    :return: distincted records


    .. doctest::
    
        >>> buyTable = [
        ...     {'name': 'Xiaomi', 'product': 'battery', 'amount':7},
        ...     {'name': 'Hong', 'product': 'keyboard', 'amount':1},
        ...     {'name': 'Lyn', 'product': 'cleaner', 'amount':5},
        ...     {'name': 'Hong', 'product': 'monitor', 'amount':1},
        ...     {'name': 'Hong', 'product': 'mouse', 'amount':3},
        ...     {'name': 'Lyn', 'product': 'mouse', 'amount':1},
        ...     {'name': 'Unknown', 'product': 'keyboard', 'amount':1},
        ...     {'name': 'Lee', 'product': 'hardcase', 'amount':2},
        ...     {'name': 'Lee', 'product': 'keycover', 'amount':2},
        ...     {'name': 'Yuki', 'product': 'manual', 'amount':1},
        ...     {'name': 'Xiaomi', 'product': 'cable', 'amount':1},
        ...     {'name': 'anonymous', 'product': 'adopter', 'amount':2},
        ...     {'name': 'Park', 'product': 'battery', 'amount':2},
        ...     {'name': 'Hong', 'product': 'cleaner', 'amount':3},
        ...     {'name': 'Smith', 'product': 'mouse', 'amount':1},
        ... ]

        >>> # maintains the existing item order if keep_first is True
        >>> records = listorm.distinct(buyTable, 'name')
        >>> for row in records: row
        {'name': 'Xiaomi', 'product': 'battery', 'amount': 7}
        {'name': 'Hong', 'product': 'keyboard', 'amount': 1}
        {'name': 'Lyn', 'product': 'cleaner', 'amount': 5}
        {'name': 'Unknown', 'product': 'keyboard', 'amount': 1}
        {'name': 'Lee', 'product': 'hardcase', 'amount': 2}
        {'name': 'Yuki', 'product': 'manual', 'amount': 1}
        {'name': 'anonymous', 'product': 'adopter', 'amount': 2}
        {'name': 'Park', 'product': 'battery', 'amount': 2}
        {'name': 'Smith', 'product': 'mouse', 'amount': 1}

        # retrive duplicated items occured at last if keep_first is False
        >>> records = listorm.distinct(buyTable, 'name', keep_first=False)
        >>> for row in records: row
        {'name': 'Lyn', 'product': 'mouse', 'amount': 1}
        {'name': 'Unknown', 'product': 'keyboard', 'amount': 1}
        {'name': 'Lee', 'product': 'keycover', 'amount': 2}
        {'name': 'Yuki', 'product': 'manual', 'amount': 1}
        {'name': 'Xiaomi', 'product': 'cable', 'amount': 1}
        {'name': 'anonymous', 'product': 'adopter', 'amount': 2}
        {'name': 'Park', 'product': 'battery', 'amount': 2}
        {'name': 'Hong', 'product': 'cleaner', 'amount': 3}
        {'name': 'Smith', 'product': 'mouse', 'amount': 1}

        # distinct by unique together by name, amount
        >>> records = listorm.distinct(buyTable, 'name', 'amount')
        >>> for row in records: row
        {'name': 'Xiaomi', 'product': 'battery', 'amount': 7}
        {'name': 'Hong', 'product': 'keyboard', 'amount': 1}
        {'name': 'Lyn', 'product': 'cleaner', 'amount': 5}
        {'name': 'Hong', 'product': 'mouse', 'amount': 3}
        {'name': 'Lyn', 'product': 'mouse', 'amount': 1}
        {'name': 'Unknown', 'product': 'keyboard', 'amount': 1}
        {'name': 'Lee', 'product': 'hardcase', 'amount': 2}
        {'name': 'Yuki', 'product': 'manual', 'amount': 1}
        {'name': 'Xiaomi', 'product': 'cable', 'amount': 1}
        {'name': 'anonymous', 'product': 'adopter', 'amount': 2}
        {'name': 'Park', 'product': 'battery', 'amount': 2}
        {'name': 'Smith', 'product': 'mouse', 'amount': 1}

        >>> # if singles is True, retrive the item only the name exist once
        >>> records = listorm.distinct(buyTable, 'name', singles=True)
        >>> for row in records: row
        {'name': 'Unknown', 'product': 'keyboard', 'amount': 1}
        {'name': 'Yuki', 'product': 'manual', 'amount': 1}
        {'name': 'anonymous', 'product': 'adopter', 'amount': 2}
        {'name': 'Park', 'product': 'battery', 'amount': 2}
        {'name': 'Smith', 'product': 'mouse', 'amount': 1}


    .. note::

        * distinct applies not only to records but also to lists with general values
        * Also removes duplicates while preserving order


        .. doctest::

            >>> dup_numbers = [1,1,2,3,3,4,4,5,5,5,6,7,7,7,8,9]
            >>> listorm.distinct(dup_numbers)
            [1, 2, 3, 4, 5, 6, 7, 8, 9]
            >>> listorm.distinct(dup_numbers, singles=True)
            [2, 6, 8, 9]

            >>> fruitbasket = [
            ...   'apple', 'apple', 'banana', 'cherry', 'cherry', 'cherry', 'mango', 'orange', 'orange', 'grape', 'kiwi', 'kiwi'
            ... ]
            >>> listorm.distinct(fruitbasket)
            ['apple', 'banana', 'cherry', 'mango', 'orange', 'grape', 'kiwi']
            >>> listorm.distinct(fruitbasket, singles=True)
            ['banana', 'mango', 'grape']


    '''
    if not keep_first:
        records = list(reversed(records))
    duplicates = {}
    for row in records:
        if isinstance(row, abc.Mapping):
            values = asvalues(row, keys)
        else:
            values = row
        duplicates.setdefault(values, []).append(row)
    
    distincts = []
    for dct, rows in duplicates.items():
        if singles and len(rows) > 1:
            continue
        distincts.append(rows[0])
    
    if not keep_first:
        distincts = list(reversed(distincts))
    return distincts


def asgroup(records:list[dict], keys:list, with_pos:bool=False) -> tuple[list, dict[str, list]]:
    grouped = {}
    for p, row in enumerate(records):
        values = asvalues(row, keys)    
        grouped.setdefault(values, []).append((p, row) if with_pos else row)
    return grouped


def aggregate(grouped:dict[str, list[dict]], keys:list, aggset:dict, aliases:dict=None, groupset_name:str=None) -> list[dict]:
    aliases = aliases or {}
    aggregated = []
    for _, rows in grouped.items():
        agged = asselect(rows[0], keys)
        for key, apply in aggset.items():
            values_list = values(rows, [key])
            agg = apply(values_list)
            alias = aliases.get(key, key)
            agged[alias] = agg
        if groupset_name:
            agged[groupset_name] = rows
        aggregated.append(agged)
    return aggregated


@reduce_args('keys')
@reduce_kwargs('aggset')
def groupby(records:list[dict], keys:list, *, aggset:dict, renames:dict=None, groupset_name:str=None) -> list[dict]:
    '''grouping by keys and aggregate values

    :param records: a list of dict items
    :param keys: keys for grouping
    :param aggset: aggregate key: function map
    :param renames: the name of aggregated key, if None, as original keyname, default None
    :param groupset_name: if specified, grouped items are dangling, default None
    :return: newly aggregated records


    .. doctest::

        >>> # get gender numbers by location
        >>> grouped = listorm.groupby(userTable,
        ...     'location', 'gender', #1. grouping
        ...      gender=len,  #2 caclulating
        ...      renames={'gender': 'gender_count'} #3 retriving as other keyname
        ... )
        >>> for row in grouped: row
        {'location': 'Korea', 'gender': 'M', 'gender_count': 2}
        {'location': 'USA', 'gender': 'M', 'gender_count': 2}
        {'location': 'China', 'gender': 'F', 'gender_count': 1}
        {'location': 'China', 'gender': 'M', 'gender_count': 1}
        {'location': 'Korea', 'gender': 'F', 'gender_count': 1}

        >>> # get average of age by gender
        >>> def calc_average_age(values):
        ...     return sum(values) / len(values)
        >>> grouped = listorm.groupby(userTable,
        ...   'gender',
        ...   age=calc_average_age,
        ...   renames={'age': 'age_avg'}
        ... )
        >>> grouped
        [{'gender': 'M', 'age_avg': 19.6}, {'gender': 'F', 'age_avg': 20.0}]

        >>> # simplify to values
        >>> listorm.values(grouped)
        [('M', 19.6), ('F', 20.0)]


        .. note::
            
            Although groupby's parameterizing is rather complicated, tt is largely divided into three parts.
            * keys
            * aggregation
            * renaming for retrieve aggregation

    '''
    grouped = asgroup(records, keys)
    agged = aggregate(grouped, keys, aggset, renames, groupset_name)
    return agged


def product(records1:list[dict], records2:list[dict]):
    for row1 in records1:
        for row2 in records2:
            row = {}
            row.update(row1)
            row.update(row2)
            yield row


@reduce_args('keys')
def set_index(records:list[dict], keys:list) -> list[tuple[tuple, dict]]:
    return [
        (asvalues(row, keys), row) for row in records
    ]

@pluralize_params('on', 'left_on', 'right_on')
def join(left:list[dict], right:list[dict], on:tuple=None, left_on:tuple=None, right_on:tuple=None, how:str='inner') -> list[dict]:
    '''merge two records

    :param left: a records
    :param right: another records
    :param on: when the field name of the key to join is the same, specify on, defaluts to None
    :param left_on: keys for join on left side records
    :param right_on: keys for join on right side records
    :param how: 'innder', 'left', 'right', 'outer', defaults to 'inner'
    :return: new merged records


    .. doctest::

        >>> # inner join by name
        >>> records = listorm.join(userTable, buyTable, on='name')
        >>> for row in records: row
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'keyboard', 'amount': 1}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'monitor', 'amount': 1}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'mouse', 'amount': 3}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'cleaner', 'amount': 3}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'cleaner', 'amount': 5}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'mouse', 'amount': 1}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'battery', 'amount': 7}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'cable', 'amount': 1}
        {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'product': 'battery', 'amount': 2}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'product': 'mouse', 'amount': 1}
        {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'hardcase', 'amount': 2}
        {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'keycover', 'amount': 2}

        >>> # left join, Charse has no buy item
        >>> records = listorm.join(userTable, buyTable, on='name', how='left')
        >>> for row in records: row
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'keyboard', 'amount': 1}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'monitor', 'amount': 1}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'mouse', 'amount': 3}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'cleaner', 'amount': 3}
        {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'cleaner', 'amount': 5}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'mouse', 'amount': 1}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'battery', 'amount': 7}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'cable', 'amount': 1}
        {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'product': 'battery', 'amount': 2}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'product': 'mouse', 'amount': 1}
        {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'hardcase', 'amount': 2}
        {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'keycover', 'amount': 2}

        >>> # merged records has missing values so filled missing as None
        >>> for row in listorm.fillmissed(records): row
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'keyboard', 'amount': 1}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'monitor', 'amount': 1}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'mouse', 'amount': 3}
        {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'cleaner', 'amount': 3}
        {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA', 'product': None, 'amount': None}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'cleaner', 'amount': 5}
        {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'mouse', 'amount': 1}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'battery', 'amount': 7}
        {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'cable', 'amount': 1}
        {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'product': 'battery', 'amount': 2}
        {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'product': 'mouse', 'amount': 1}
        {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'hardcase', 'amount': 2}
        {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'keycover', 'amount': 2}


    '''
    if on:
        left_on = right_on = on
    elif not all([left_on, right_on]):
        raise ValueError('left_on and rignt_on must be specified')

    left_group = asgroup(left, left_on)
    right_group = asgroup(right, right_on)

    leftset = left_group.keys()
    rightset = right_group.keys()
    intersection = leftset & rightset

    leftkeys = [key for key in left_group if key in leftset]
    rightkeys = [key for key in right_group if key in rightset]
    unionkeys = distinct(leftkeys+rightkeys)

    if how == 'left':
        joinkeys = leftkeys
    elif how == 'right':
        joinkeys = rightkeys
    elif how == 'inner':
        joinkeys = [key for key in unionkeys if key in intersection]
    elif how == 'outer':
        joinkeys = unionkeys
    else:
        raise ValueError

    joined = []
    for jkey in joinkeys:
        lrows = left_group.get(jkey, [{}])
        rrows = right_group.get(jkey, [{}])
        products = (
            (lrow, rrow)
            for lrow in lrows
            for rrow in rrows
        )
        
        rows = []
        for lrow, rrow in products:
            row = {}
            row.update(lrow)
            row.update(rrow)
            rows.append(row)
        joined += rows
    return joined


@reduce_args('keys')
def values_count(records:list[dict], keys:list):
    '''get values count by keys

    :param records: a list contains dict items
    :param keys: key for counted values
    :return: dict has values count


    .. doctest::

        >>> listorm.values_count(buyTable, 'product')
        {'battery': 2, 'keyboard': 2, 'cleaner': 2, 'monitor': 1, 'mouse': 3, 'hardcase': 1, 'keycover': 1, 'manual': 1, 'cable': 1, 'adopter': 1}

    '''

    counter = {}
    for row in records:
        value = asvalues(row, keys)
        counter.setdefault(value, 0)
        counter[value] += 1   
    return counter


@reduce_kwargs('formats')
def set_number_format(records:list[dict], *, formats:dict=None):
    '''change number formats as default examples

    :param records: a list contains dict items
    :param formats: number format examples, defaults to None
    :return: number formatted records


    .. doctest::

        >>> numbers = [
        ...     {
        ...       'flt': 0.5, 'string': '123', 'string_float': '123.5', 'int': 412, 'string_int': '5123', 'blabla': 'what?',  
        ...     }
        ... ]

        >>> # key name and example of types to change, if faild to change, example value will be default value
        >>> listorm.set_number_format(numbers, flt='', string=0.0, string_float=0, int='', string_int=0, blabla=0)
        [{'flt': '0.5', 'string': 123.0, 'string_float': 123, 'int': '412', 'string_int': 5123, 'blabla': 0}]

    '''
    return [
        asnumformat(row, formats) for row in records
    ]

@reduce_args('keys')
def is_unique(records:list[dict], keys:list):
    '''check unique for values of keys


    '''
    counter = values_count(records, keys)
    return max(counter.values(), default=1) < 2


@pluralize_params('pk')
def diff(records1:list[dict], records2:list[dict], pk:tuple):    
    '''compare two records about added, deleted and updated on common columns

    :param records1: a list object as records
    :param records2: another list objects as records
    :param pk: unique keys for both of two records
    :raises ValueError: when empty records passed
    :raises UniqueConstraintError: when values for pk of reocords, is not unique
    :return: namedtuple objects as Changes


    .. doctest::


        >>> before = [
        ...    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
        ...    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
        ...    {'name': 'ohmyboss', 'gender': 'F', 'age': 17, 'location': 'USA'},
        ...    {'name': 'Lyn', 'gender': 'F', 'age': 29, 'location': 'China'},
        ... ]
        >>> after = [
        ...    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
        ...    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
        ...    {'name': 'ohmyboss', 'gender': 'M', 'age': 17, 'location': 'Canada'},
        ...    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
        ... ]


        >>> changes = listorm.diff(before, after, 'name')

        >>> changes.added
        [Added(pk='Xiaomi', rows={'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'}), Added(pk='Park', rows={'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'})]

        >>> changes.deleted
        [Deleted(pk='Hong', rows={'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'}), Deleted(pk='Charse', rows={'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'})]

        >>> changes.updated
        [Updated(pk='ohmyboss', before={'name': 'ohmyboss', 'gender': 'F', 'age': 17, 'location': 'USA'}, after={'name': 'ohmyboss', 'gender': 'M', 'age': 17, 'location': 'Canada'}, where=['gender', 'location']), Updated(pk='Lyn', before={'name': 'Lyn', 'gender': 'F', 'age': 29, 'location': 'China'}, after={'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'}, where=['age'])]

        >>> # usage of updated
        >>> for change in changes.updated:
        ...     print(change.pk)
        ...     for key in change.where:
        ...         print('where:', key)
        ...         print('  before->after:',change.before[key], '->', change.after[key])
        ...     print('------------------------------')
        ohmyboss
        where: gender
          before->after: F -> M
        where: location
          before->after: USA -> Canada
        ------------------------------
        Lyn
        where: age
          before->after: 29 -> 28
        ------------------------------

    '''
    if not records1 or not records2:
        raise ValueError('not allowed empthy records')

    if not all([is_unique(records1, pk), is_unique(records2, pk)]):
        raise UniqueConstraintError
    
    beforeset = asgroup(records1, pk)
    afterset = asgroup(records2, pk)

    comparisonkeys = distinct([*beforeset, *afterset])

    Added = namedtuple('Added', 'pk rows')
    Deleted = namedtuple('Deleted', 'pk rows')
    Updated = namedtuple('Updated', 'pk before after where')
    Changes = namedtuple('Changes', 'added deleted updated')

    added = []
    deleted = []
    updated = []
    for key in comparisonkeys:
        before = beforeset.get(key)
        after = afterset.get(key)
        if before and not after:
            deleted.append(Deleted(key, before[0]))
        elif after and not before:
            added.append(Added(key, after[0]))
        else:
            before, after = before[0], after[0]
            diff = asdiff(before, after)
            if diff:
                updated.append(Updated(key, before, after, diff))
    return Changes(added, deleted, updated)








'''
Class Based API for records
--------------------------------
'''


from .base import BaseList
from .shortcuts import ShortCutMixin
from ..exceptions import *
from ..api import *
from ..utils import reduce_args, reduce_kwargs, pluralize_params



class Listorm(ShortCutMixin, BaseList):
    '''wrapper class for 

    :param records: list as records
    :param uniques: if specified, it check unique constraint, defaults to None
    :param fill_missed: if True fill missing values of rows, defaults to True
    :param fill_value: if specified, fill missing as specifed value, defaults to None

    '''

    def filter(self, where:callable):
        records = [
            row for row in self if reduce_where(row, where)
        ]
        return Listorm(
            records,
            uniques=self.uniques, fill_missed=False
        )


    @reduce_args('columns')
    def select(self, columns:list, *, excludes:list=None, where:callable=None):
        '''Retrieves the item of the specified item

        :param columns: keys for selecting columns
        :param excludes: keys for excluding columns, defaults to None
        :param where: callback for filtering records, defaults to None
        :return: Listorm

        
        .. doctest::

            >>> from listorm import Listorm
            >>> userTable = [
            ...    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
            ...    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
            ...    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
            ...    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
            ...    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
            ...    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
            ...    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'},
            ... ]

            ls = Listorm(userTable)
            ls.select('name', 'age', 'location', where=lambda age: age > 20).print()
            {'name': 'Lyn', 'age': 28, 'location': 'China'}
            {'name': 'Park', 'age': 29, 'location': 'Korea'}


        '''
        return Listorm(
            select(self, columns, excludes=excludes, where=where), 
            fill_missed=False
        )


    @reduce_args('columns')
    def drop_column(self, columns:list):
        '''delete columns


        .. doctest::

            >>> from listorm import Listorm

            >>> userTable = [
            ...    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
            ...    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
            ...    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
            ...    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
            ...    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
            ...    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
            ...    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'},
            ... ]

            >>> ls = Listorm(userTable)
            >>> ls.drop_column('age', 'location').print()
            {'name': 'Hong', 'gender': 'M'}
            {'name': 'Charse', 'gender': 'M'}
            {'name': 'Lyn', 'gender': 'F'}
            {'name': 'Xiaomi', 'gender': 'M'}
            {'name': 'Park', 'gender': 'M'}
            {'name': 'Smith', 'gender': 'M'}
            {'name': 'Lee', 'gender': 'F'}


        :param columns: column name for delete
        :return: Listorm
        '''
        return self.select(excludes=columns)
    

    @reduce_kwargs('keymap')
    def add_column(self, keymap:dict):
        '''extend row as new columns

        :param keymap: new column: function mapping
        :return: Listorm

        
        .. doctest::

            >>> from listorm import Listorm

            >>> userTable = [
            ...    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
            ...    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
            ...    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
            ...    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
            ...    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
            ...    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
            ...    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'},
            ... ]

            >>> ls = Listorm(userTable)

            >>> ls.add_column(
            ...    summary=lambda name, gender, age: f"{name}/{gender}/{age}"
            ... ).print()
            {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'summary': 'Hong/M/18'}
            {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA', 'summary': 'Charse/M/19'}
            {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'summary': 'Lyn/F/28'}
            {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'summary': 'Xiaomi/M/15'}
            {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'summary': 'Park/M/29'}
            {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'summary': 'Smith/M/17'}
            {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'summary': 'Lee/F/12'}

        '''
        return Listorm(
            add_column(self, keymap=keymap),
            **self.as_kwargs(fill_missed=False)
        )


    @reduce_kwargs('renamemap')
    def rename(self, renamemap:dict):
        '''change columns name

        :param renamemap: old:new mapping of column names
        :return: Listorm

        .. doctest::


            >>> from listorm import Listorm
            >>> ls = Listorm(userTable)

            >>> ls.rename(gender='sex', location='country').print()
            {'name': 'Hong', 'sex': 'M', 'age': 18, 'country': 'Korea'}
            {'name': 'Charse', 'sex': 'M', 'age': 19, 'country': 'USA'}
            {'name': 'Lyn', 'sex': 'F', 'age': 28, 'country': 'China'}
            {'name': 'Xiaomi', 'sex': 'M', 'age': 15, 'country': 'China'}
            {'name': 'Park', 'sex': 'M', 'age': 29, 'country': 'Korea'}
            {'name': 'Smith', 'sex': 'M', 'age': 17, 'country': 'USA'}
            {'name': 'Lee', 'sex': 'F', 'age': 12, 'country': 'Korea'}

            >>> # if key name has to be containing spaces
            >>> ls.rename(renamemap={'age': 'age of ultron' }).print()
            {'name': 'Hong', 'gender': 'M', 'age of ultron': 18, 'location': 'Korea'}
            {'name': 'Charse', 'gender': 'M', 'age of ultron': 19, 'location': 'USA'}
            {'name': 'Lyn', 'gender': 'F', 'age of ultron': 28, 'location': 'China'}
            {'name': 'Xiaomi', 'gender': 'M', 'age of ultron': 15, 'location': 'China'}
            {'name': 'Park', 'gender': 'M', 'age of ultron': 29, 'location': 'Korea'}
            {'name': 'Smith', 'gender': 'M', 'age of ultron': 17, 'location': 'USA'}
            {'name': 'Lee', 'gender': 'F', 'age of ultron': 12, 'location': 'Korea'}

        '''
        return Listorm(rename(self, renamemap=renamemap), fill_missed=False)


    @reduce_kwargs('updatemap')
    def update(self, updatemap:dict, *, where:callable=None):
        '''update row values

        :param updatemap: column: value|function map for update values
        :param where: specifying what to update, defaults to None
        :return: Listorm


        .. doctest::

            >>> ls.update(
            ...    age=lambda age: age + 1,
            ...    where=lambda location: location.lower() in ['korea', 'china']
            ... ).print()
            {'name': 'Hong', 'gender': 'M', 'age': 19, 'location': 'Korea'}
            {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'}
            {'name': 'Lyn', 'gender': 'F', 'age': 29, 'location': 'China'}
            {'name': 'Xiaomi', 'gender': 'M', 'age': 16, 'location': 'China'}
            {'name': 'Park', 'gender': 'M', 'age': 30, 'location': 'Korea'}
            {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}
            {'name': 'Lee', 'gender': 'F', 'age': 13, 'location': 'Korea'}

        '''
        return Listorm(
            update(self, updatemap=updatemap, where=where),
            fill_missed=False
        )
    

    @reduce_args('columns')
    def values(self, columns:list, *, flat_one=True):
        '''extract values from row

        :param columns: column names to extract values
        :param flat_one: If True, when the return value is a single value, the value is returned without being included in the list, defaults to True, defaults to True
        :return: list


        .. doctest::

            >>> values = ls.values('name', 'gender', 'location')
            >>> for v in values: print(v)
            ('Hong', 'M', 'Korea')
            ('Charse', 'M', 'USA')
            ('Lyn', 'F', 'China')
            ('Xiaomi', 'M', 'China')
            ('Park', 'M', 'Korea')
            ('Smith', 'M', 'USA')
            ('Lee', 'F', 'Korea')

            >>> # When there is only one column, a flatten single list is returned
            >>> ls.values('name')
            ['Hong', 'Charse', 'Lyn', 'Xiaomi', 'Park', 'Smith', 'Lee']

            >>> # but, flat_one = False when consistency is required
            >>> ls.values('name', flat_one=False)
            [('Hong',), ('Charse',), ('Lyn',), ('Xiaomi',), ('Park',), ('Smith',), ('Lee',)]

        '''
        return values(self, columns, flat_one=flat_one)


    @reduce_args('columns')
    def distinct(self, columns:list, *, keep_first:bool=True, singles:bool=False):
        '''remove duplicated rows by columns

        :param columns: column names for check duplicated values
        :param keep_first: If True, in case of duplicate occurrence select the first item that appears, else last, defaults to True
        :param singles: If True, eliminate rows that have been duplicated. That is, only items that were unique are left. defaults to False, defaults to False
        :return: Listorm


        .. doctest::

            >>> buyTable = [
            ...     {'name': 'Xiaomi', 'product': 'battery', 'amount':7},
            ...     {'name': 'Hong', 'product': 'keyboard', 'amount':1},
            ...     {'name': 'Lyn', 'product': 'cleaner', 'amount':5},
            ...     {'name': 'Hong', 'product': 'monitor', 'amount':1},
            ...     {'name': 'Hong', 'product': 'mouse', 'amount':3},
            ...     {'name': 'Lyn', 'product': 'mouse', 'amount':1},
            ...     {'name': 'Unknown', 'product': 'keyboard', 'amount':1},
            ...     {'name': 'Lee', 'product': 'hardcase', 'amount':2},
            ...     {'name': 'Lee', 'product': 'keycover', 'amount':2},
            ...     {'name': 'Yuki', 'product': 'manual', 'amount':1},
            ...     {'name': 'Xiaomi', 'product': 'cable', 'amount':1},
            ...     {'name': 'anonymous', 'product': 'adopter', 'amount':2},
            ...     {'name': 'Park', 'product': 'battery', 'amount':2},
            ...     {'name': 'Hong', 'product': 'cleaner', 'amount':3},
            ...     {'name': 'Smith', 'product': 'mouse', 'amount':1},
            ... ]

            >>> len(buyTable)
            15

            >>> lsb = Listorm(buyTable)

            >>> lsb.distinct('name', 'amount').count
            12

            >>> lsb.distinct('name').print()
            {'name': 'Xiaomi', 'product': 'battery', 'amount': 7}
            {'name': 'Hong', 'product': 'keyboard', 'amount': 1}
            {'name': 'Lyn', 'product': 'cleaner', 'amount': 5}
            {'name': 'Unknown', 'product': 'keyboard', 'amount': 1}
            {'name': 'Lee', 'product': 'hardcase', 'amount': 2}
            {'name': 'Yuki', 'product': 'manual', 'amount': 1}
            {'name': 'anonymous', 'product': 'adopter', 'amount': 2}
            {'name': 'Park', 'product': 'battery', 'amount': 2}
            {'name': 'Smith', 'product': 'mouse', 'amount': 1}


            # distinct by unique together by name, amount
            >>> lsb.distinct('name', 'amount').print()
            {'name': 'Xiaomi', 'product': 'battery', 'amount': 7}
            {'name': 'Hong', 'product': 'keyboard', 'amount': 1}
            {'name': 'Lyn', 'product': 'cleaner', 'amount': 5}
            {'name': 'Hong', 'product': 'mouse', 'amount': 3}
            {'name': 'Lyn', 'product': 'mouse', 'amount': 1}
            {'name': 'Unknown', 'product': 'keyboard', 'amount': 1}
            {'name': 'Lee', 'product': 'hardcase', 'amount': 2}
            {'name': 'Yuki', 'product': 'manual', 'amount': 1}
            {'name': 'Xiaomi', 'product': 'cable', 'amount': 1}
            {'name': 'anonymous', 'product': 'adopter', 'amount': 2}
            {'name': 'Park', 'product': 'battery', 'amount': 2}
            {'name': 'Smith', 'product': 'mouse', 'amount': 1}

            >>> # if singles is True, retrive the item only the name exist once
            >>> lsb.distinct('name', singles=True).print()
            {'name': 'Unknown', 'product': 'keyboard', 'amount': 1}
            {'name': 'Yuki', 'product': 'manual', 'amount': 1}
            {'name': 'anonymous', 'product': 'adopter', 'amount': 2}
            {'name': 'Park', 'product': 'battery', 'amount': 2}
            {'name': 'Smith', 'product': 'mouse', 'amount': 1}

        '''
        return Listorm(distinct(self, columns, keep_first=keep_first, singles=singles), fill_missed=False)


    @reduce_args('sortkeys')
    def orderby(self, sortkeys:list):
        '''sort rows 

        :param sortkeys: column names or function for make sort keys
        :return: Listorm

        .. doctest::

            >>> # complex key sorting: location as ascending and by age decending
            >>> ls.orderby(userTable, 'location', '-age').print()
            {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'}
            {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'}
            {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'}
            {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'}
            {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'}
            {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'}
            {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}

            >>> # ordering by callback
            >>> # order by first digit of age
            >>> ls.orderby(userTable, lambda age: str(age)[-1]).print()
            {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'}
            {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'}
            {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'}
            {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'}
            {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'}
            {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'}
            {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'}

        .. note::

            If you are familiar with django it looks like it

        '''
        return Listorm(orderby(self, sortkeys), **self.as_kwargs(fill_missed=False))


    @reduce_args('columns')
    @reduce_kwargs('aggset')
    def groupby(self, columns:list, *, aggset:dict, renames:dict=None, groupset_name:str=None):
        '''grouping rows and aggregate valeus

        :param columns: column names for grouping
        :param aggset: aggregate column: function map
        :param renames: the name of aggregated columns, if None, as original column name, default None
        :param groupset_name: if specified, grouped items are dangling, default None
        :return: Listorm


        .. doctest::

            >>> # get gender numbers by location
            >>> ls.groupby(
            ...     'location', 'gender', #1. grouping
            ...      gender=len,  #2 caclulating
            ...      renames={'gender': 'gender_count'} #3 retriving as other keyname
            ... ).print()
            {'location': 'Korea', 'gender': 'M', 'gender_count': 2}
            {'location': 'USA', 'gender': 'M', 'gender_count': 2}
            {'location': 'China', 'gender': 'F', 'gender_count': 1}
            {'location': 'China', 'gender': 'M', 'gender_count': 1}
            {'location': 'Korea', 'gender': 'F', 'gender_count': 1}

            >>> # get average of age by gender
            >>> def calc_average_age(values):
            ...     return sum(values) / len(values)

            >>> ls.groupby(
            ...   'gender',
            ...   age=calc_average_age,
            ...   renames={'age': 'age_avg'}
            ... )
            [{'gender': 'M', 'age_avg': 19.6}, {'gender': 'F', 'age_avg': 20.0}]


            .. note::
                
                Although groupby's parameterizing is rather complicated, tt is largely divided into three parts.
                * keys
                * aggregation
                * renaming for retrieve aggregation


        '''
        return Listorm(
            groupby(self, columns, aggset=aggset, renames=renames, groupset_name=groupset_name)
        )


    @pluralize_params('left_on', 'right_on')
    def join(self, other, left_on:None, right_on=None, how:str='inner'):
        '''merge other records

        :param other: another Listorm object
        :param left_on: columns for join on left side records
        :param right_on: columns for join on right side records
        :param how: how to join 'inner', 'left', 'right', 'outer', defaults to 'inner'
        :return: Listorm


        '''
        left_on = left_on or self.uniques        
        
        if not right_on:
            if isinstance(other, Listorm):
                right_on = other.uniques or left_on
            else:
                right_on = left_on
        if not all((left_on, right_on)):
            raise JoinKeyDoesNotExists('on:{}, right_on:{} must be specified'.format(left_on, right_on))
        return Listorm(
                join(self, other, None, left_on, right_on, how=how),
                fill_value=self.fill_value
            )

    



    


from collections import abc

from ..utils import get_argcounts, filer_kwargs



def reduce_args_count(app, *args):
    if callable(app):
        args_count = get_argcounts(app)
        if args_count > 2:
            raise ValueError("{} can have no more than 2 args.".format(app))
        return app(*args[:args_count])
    return app


def reduce_where(item, where):
    if where is None:
        return True
    
    if not callable(where):
        raise ValueError('where must be callable')
    
    try:
        return where(**item)
    except TypeError:
        kwargs = filer_kwargs(item, where)
        return where(**kwargs)



def reduce_callback(item, key=None, app=None):
    updated = {}
    if not callable(app):
        if isinstance(app, abc.Mapping):
            value = app.get(item[key], item[key])
        else:
            value = app
    else:
        try:
            kwargs = filer_kwargs(item, app)
        except ValueError:
            value = app(item[key])
        else:
            try:
                value = app(**item)
            except TypeError:
                if kwargs:
                    value = app(**kwargs)
                else:
                    value = app(item[key])
    if key is None:
        return value

    updated[key] = value
    return updated



from collections import UserList

from ..api.records import *
from ..api.extensions import *
from ..utils import tuplize
from ..exceptions import UniqueConstraintError



class BaseList(UserList):


    def __init__(self, records:list, uniques:tuple=None, fill_missed=True,  fill_value=None):
        self.uniques = tuplize(uniques)
        self.fill_value = fill_value
        self.fill_missed = fill_missed
        if self.uniques:
            records = self.check_for_uniques(records)
        if self.fill_missed:
            records = self.normalize_records(records, fill_value)
        super().__init__(records)

    def as_kwargs(self, **updates):
        default = asselect(self.__dict__, ['uniques', 'fill_value', 'fill_missed'])
        default.update(updates)
        return default

    def check_for_uniques(self, records):
        uniqueset = set(self.uniques)
        seen = {}
        for row in records:
            missing = uniqueset - row.keys()
            if missing:
                missing = ', '.join(missing)
                raise UniqueConstraintError(
                    "The key(s) of {} does not exist in the {} for unique constraint validation".format(missing, row)
                )
            values = asvalues(row, self.uniques)
            exists = seen.get(values)
            if exists:
                raise UniqueConstraintError(
                    "\nUnique Constraint Failed in.\n{}\n{}\non:{}".format(exists, row, self.uniques)
                )
            seen[values] = row
            yield row

    def normalize_records(self, records, value):
        return fillmissed(records, value)
    
    @classmethod
    def from_csv(cls, file, uniques=None, encoding='utf-8'):
        return cls(read_csv(file, encoding=encoding), uniques)
    
    def print(self, nrow=25, **print_kwargs):
        for i, row in enumerate(self):
            if i == nrow:
                break
            print(row, **print_kwargs)


import pytest

from listorm import Listorm, read_csv, read_excel
from listorm.exceptions import UniqueConstraintError
from .samples import *



test_init_arg_cases = [
    (userTable, None, False, None, userTable),
    (userTable_missing_values, None, False, None, userTable_missing_values),
    (userTable_missing_values, None, False, None, userTable_missing_values),
    (userTable_missing_values, None, True, 'undefined', userTable_missing_values_filling_with_undefined),
    (userTable, ('location', 'gender',), False, None, userTable),
    (userTable_missing_values, 'name', False, None, userTable_missing_values),
]
@pytest.mark.parametrize('records, uniques, fill_missed, fill_value, results', test_init_arg_cases)
def test_init(records, uniques, fill_missed, fill_value, results):
    if not uniques:
        lst = Listorm(records,uniques,fill_missed,fill_value)
        assert lst == results

    if uniques:
        with pytest.raises(UniqueConstraintError):
            lst = Listorm(records,uniques,fill_missed,fill_value)


test_select_cases = [
    (userTable, 'name', ['location'], None, userTable_select_name_exclude_location),
    (userTable, [], None, lambda age: age > 20, userTable_age_over_20),
    (userTable, ['name', 'gender', 'location' ], None, lambda age: age > 20, userTable_select_name_gender_location_where_age_gt_20),
]
@pytest.mark.parametrize('records, columns, excludes, where, results', test_select_cases)
def test_select(records, columns, excludes, where, results):
    lst = Listorm(records).select(columns, excludes=excludes, where=where)
    assert lst == results


test_drop_column_cases = [
    (userTable, ['age'], userTable_dropped_age),
    (userTable, ['gender', 'location'], userTable_dropped_gender_location),
]
@pytest.mark.parametrize('records, columns, results', test_drop_column_cases)
def test_drop_column(records, columns, results):
    lst = Listorm(records).drop_column(columns)
    assert lst == results


test_add_column_cases = [
    (userTable, {'key': 'value'}, userTable_column_added_value),
    (userTable, {'gender/age': lambda gender, age: '{}/{}'.format(gender, age)}, userTable_column_added_gender_age_concated)
]
@pytest.mark.parametrize('records, column_mapset, results', test_add_column_cases)
def test_add_column(records, column_mapset, results):
    lst = Listorm(records).add_column(**column_mapset)
    assert lst == results


test_rename_cases = [
    (userTable, {'location': 'address'}, userTable_renamed_location2address),
    (userTable, {'location': 'address', 'gender': 'sex'}, userTable_renamed_location2address_gender2sex),
]
@pytest.mark.parametrize('records, renamemap, results', test_rename_cases)
def test_rename(records, renamemap, results):
    lst = Listorm(records).rename(**renamemap)
    assert lst == results


test_update_cases = [
    (userTable, {'name': str.upper}, None, userTable_updated_name_upper),
    (userTable, {'name': str.upper}, lambda age: age > 20, userTable_updated_name_upper_where_age_gt_20),
    (userTable, {'gender': lambda gender, age: "{}/{}".format(gender, age) }, None, userTable_updated_gender_concat_with_age),
]
@pytest.mark.parametrize('records, updatemap, where, results', test_update_cases)
def test_update(records, updatemap, where, results):
    lst = Listorm(records).update(**updatemap, where=where)
    assert lst == results


test_values_cases = [
    (userTable, ['name'], True, userTable_values_by_name),
    (userTable, ['name'], False, userTable_values_by_name_flat_flase),
    (userTable, ['name', 'gender'], True, userTable_values_by_name_and_gender),
]
@pytest.mark.parametrize('records, columns, flat_one, results', test_values_cases)
def test_values(records, columns, flat_one, results):
    lst = Listorm(records).values(columns, flat_one=flat_one)
    assert lst == results


test_distinct_cases = [
    (buyTable, ['name'], True, False, buyTable_distinct_by_name_keep_first),
    (buyTable, ['name'], False, False, buyTable_distinct_by_name_keep_last),
    (buyTable, ['name'], True, True, buyTable_distinct_by_name_singles),
]
@pytest.mark.parametrize('records, columns, first, singles, results', test_distinct_cases)
def test_distinct(records, columns, first, singles, results):
    assert results == Listorm(records).distinct(columns, keep_first=first, singles=singles)


test_orderby_cases = [
    (userTable, ['location', '-age'], userTable_orderby_location_age_desc)
]
@pytest.mark.parametrize('records, sortkeys, results', test_orderby_cases)
def test_orderby(records, sortkeys, results):
    assert results == Listorm(records).orderby(sortkeys)

test_groupby_cases = [
    (userTable, ['location', 'gender'], {'gender': len, 'age': lambda vals: sum(vals)/len(vals)}, {'gender': 'count', 'age': 'avg_age'}, userTable_groupby_location_gender)
]
@pytest.mark.parametrize('records, columns, aggset, renames, results', test_groupby_cases)
def test_groupby(records, columns, aggset, renames, results):
    assert results == Listorm(records).groupby(columns, aggset=aggset, renames=renames)


test_join_cases = [
    (userTable, buyTable, 'name', None, 'inner', user_buy_table_inner_joined),
    (userTable, buyTable, 'name', None, 'left', user_buy_table_left_joined),
    (userTable, buyTable, 'name', None, 'right', user_buy_table_right_joined),
    (userTable, buyTable, 'name', None, 'outer', user_buy_table_outer_joined),
]
@pytest.mark.parametrize('records, other, on, right_on, how, results', test_join_cases)
def test_join(records, other, on, right_on, how, results):
    lst = Listorm(records, fill_value= '') \
        .join(other, on, right_on, how) \
        .orderby('name', 'product') \
        .set_number_type(amount=0)
    for rec, l in zip(results, lst):
        assert rec == l


test_write_and_read_csv_cases = [
    (userTable, 'user_table_write_test.csv', userTable),
    (userTable_missing_values, 'user_table_write_missing_values_test.csv', userTable_missing_values_filling_with_undefined),
    (userTable_has_one_column_name, 'user_table_write_has_oncolumn_test.csv', userTable_has_one_column_name)
]
@pytest.mark.parametrize('records, file, results', test_write_and_read_csv_cases)
def test_write_and_read_csv(records, file, results):
    lstsrc = Listorm(records, fill_value='undefined')
    lstsrc.to_csv(file)
    lstdest = read_csv(file).update(
        age=int,
        where=lambda **row: row.get('age') and row['age'].isnumeric()
    )
    assert results == lstdest

test_write_and_read_excel_cases = [
    (userTable, 'user_table_write_test.xlsx', userTable),
    (userTable_missing_values, 'user_table_write_missing_values_test.xlsx', userTable_missing_values_filling_with_undefined),
    (userTable_has_one_column_name, 'user_table_write_has_oncolumn_test.xlsx', userTable_has_one_column_name)
]
@pytest.mark.parametrize('records, file, results', test_write_and_read_excel_cases)
def test_write_and_read_excel(records, file, results):
    lst = Listorm(records, fill_value='undefined')
    lst.to_excel(file)
    lstdest = read_excel(file)
    assert results == lstdest

test_min_max_cases = [
    (userTable, 'max', 'age', 29),
    (userTable, 'min', 'age', 12),
]
@pytest.mark.parametrize('records, min_max, column, results', test_min_max_cases)
def test_minmax(records, min_max, column, results):
    if min_max == 'max':
        value = Listorm(records).max(column)
    else:
        value = Listorm(records).min(column)
    assert results == value


test_values_count_cases = [
    (buyTable, 'product', buyTable_values_count_by_product)
]
@pytest.mark.parametrize('records, products, results', test_values_count_cases)
def test_values_count(records, products, results):
    lst = Listorm(records).values_count(products)
    assert results == lst


userTable_missing_values_filling_with_undefined = [
    {'age': 18, 'location': 'Korea', 'name': 'undefined', 'gender': 'undefined'},
    {'name': 'Charse', 'gender': 'M', 'location': 'USA', 'age': 'undefined'},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
    {'name': 'Xiaomi', 'gender': 'undefined', 'age': 'undefined', 'location': 'undefined'},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
    {'name': 'undefined', 'gender': 'undefined', 'age': 'undefined', 'location': 'Korea'},
]

userTable_select_name_exclude_location = [
    {'name': 'Hong'},
    {'name': 'Charse'},
    {'name': 'Lyn'},
    {'name': 'Xiaomi'},
    {'name': 'Park'},
    {'name': 'Smith'},
    {'name': 'Lee'},
]

userTable_age_over_20 = [
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
]

userTable_select_name_gender_location_where_age_gt_20 = [
    {'name': 'Lyn', 'gender': 'F', 'location': 'China'},
    {'name': 'Park', 'gender': 'M', 'location': 'Korea'},
]

userTable_dropped_age = [
    {'name': 'Hong', 'gender': 'M', 'location': 'Korea'},
    {'name': 'Charse', 'gender': 'M', 'location': 'USA'},
    {'name': 'Lyn', 'gender': 'F', 'location': 'China'},
    {'name': 'Xiaomi', 'gender': 'M', 'location': 'China'},
    {'name': 'Park', 'gender': 'M', 'location': 'Korea'},
    {'name': 'Smith', 'gender': 'M', 'location': 'USA'},
    {'name': 'Lee', 'gender': 'F', 'location': 'Korea'},
]

userTable_dropped_gender_location = [
    {'name': 'Hong', 'age': 18},
    {'name': 'Charse', 'age': 19},
    {'name': 'Lyn', 'age': 28},
    {'name': 'Xiaomi', 'age': 15},
    {'name': 'Park', 'age': 29},
    {'name': 'Smith', 'age': 17},
    {'name': 'Lee', 'age': 12},
]

userTable_column_added_gender_age_concated = [
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'gender/age': 'M/18'},
    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA', 'gender/age': 'M/19'},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'gender/age': 'F/28'},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'gender/age': 'M/15'},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'gender/age': 'M/29'},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'gender/age': 'M/17'},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'gender/age': 'F/12'},
]

userTable_column_added_value = [
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'key': 'value'},
    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA', 'key': 'value'},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'key': 'value'},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'key': 'value'},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'key': 'value'},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'key': 'value'},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'key': 'value'},
]

userTable_renamed_location2address = [
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'address': 'Korea'},
    {'name': 'Charse', 'gender': 'M', 'age': 19, 'address': 'USA'},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'address': 'China'},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'address': 'China'},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'address': 'Korea'},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'address': 'USA'},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'address': 'Korea'},
]
userTable_renamed_location2address_gender2sex = [
    {'name': 'Hong', 'sex': 'M', 'age': 18, 'address': 'Korea'},
    {'name': 'Charse', 'sex': 'M', 'age': 19, 'address': 'USA'},
    {'name': 'Lyn', 'sex': 'F', 'age': 28, 'address': 'China'},
    {'name': 'Xiaomi', 'sex': 'M', 'age': 15, 'address': 'China'},
    {'name': 'Park', 'sex': 'M', 'age': 29, 'address': 'Korea'},
    {'name': 'Smith', 'sex': 'M', 'age': 17, 'address': 'USA'},
    {'name': 'Lee', 'sex': 'F', 'age': 12, 'address': 'Korea'},
]

userTable_updated_name_upper = [
    {'name': 'HONG', 'gender': 'M', 'age': 18, 'location': 'Korea'},
    {'name': 'CHARSE', 'gender': 'M', 'age': 19, 'location': 'USA'},
    {'name': 'LYN', 'gender': 'F', 'age': 28, 'location': 'China'},
    {'name': 'XIAOMI', 'gender': 'M', 'age': 15, 'location': 'China'},
    {'name': 'PARK', 'gender': 'M', 'age': 29, 'location': 'Korea'},
    {'name': 'SMITH', 'gender': 'M', 'age': 17, 'location': 'USA'},
    {'name': 'LEE', 'gender': 'F', 'age': 12, 'location': 'Korea'},
]
userTable_updated_name_upper_where_age_gt_20 = [
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
    {'name': 'LYN', 'gender': 'F', 'age': 28, 'location': 'China'},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
    {'name': 'PARK', 'gender': 'M', 'age': 29, 'location': 'Korea'},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'},
]
userTable_updated_gender_concat_with_age = [
    {'name': 'Hong', 'gender': 'M/18', 'age': 18, 'location': 'Korea'},
    {'name': 'Charse', 'gender': 'M/19', 'age': 19, 'location': 'USA'},
    {'name': 'Lyn', 'gender': 'F/28', 'age': 28, 'location': 'China'},
    {'name': 'Xiaomi', 'gender': 'M/15', 'age': 15, 'location': 'China'},
    {'name': 'Park', 'gender': 'M/29', 'age': 29, 'location': 'Korea'},
    {'name': 'Smith', 'gender': 'M/17', 'age': 17, 'location': 'USA'},
    {'name': 'Lee', 'gender': 'F/12', 'age': 12, 'location': 'Korea'},
]

userTable_values_by_name = ['Hong','Charse','Lyn','Xiaomi','Park','Smith','Lee']
userTable_values_by_name = ['Hong','Charse','Lyn','Xiaomi','Park','Smith','Lee']
userTable_values_by_name_flat_flase = [('Hong',),('Charse',),('Lyn',),('Xiaomi',),('Park',),('Smith',),('Lee',)]
userTable_values_by_name_and_gender = [
    ("Hong", "M",),
    ("Charse", "M",),
    ("Lyn", "F",),
    ("Xiaomi", "M",),
    ("Park", "M",),
    ("Smith", "M",),
    ("Lee", "F",),
]

buyTable_distinct_by_name_keep_first = [
    {'name': 'Xiaomi', 'product': 'battery', 'amount':7},
    {'name': 'Hong', 'product': 'keyboard', 'amount':1},
    {'name': 'Lyn', 'product': 'cleaner', 'amount':5},
    {'name': 'Unknown', 'product': 'keyboard', 'amount':1},
    {'name': 'Lee', 'product': 'hardcase', 'amount':2},
    {'name': 'Yuki', 'product': 'manual', 'amount':1},
    {'name': 'anonymous', 'product': 'adopter', 'amount':2},
    {'name': 'Park', 'product': 'battery', 'amount':2},
    {'name': 'Smith', 'product': 'mouse', 'amount':1},
]
buyTable_distinct_by_name_keep_last = [
    {'name': 'Lyn', 'product': 'mouse', 'amount':1},
    {'name': 'Unknown', 'product': 'keyboard', 'amount':1},
    {'name': 'Lee', 'product': 'keycover', 'amount':2},
    {'name': 'Yuki', 'product': 'manual', 'amount':1},
    {'name': 'Xiaomi', 'product': 'cable', 'amount':1},
    {'name': 'anonymous', 'product': 'adopter', 'amount':2},
    {'name': 'Park', 'product': 'battery', 'amount':2},
    {'name': 'Hong', 'product': 'cleaner', 'amount':3},
    {'name': 'Smith', 'product': 'mouse', 'amount':1},
]
buyTable_distinct_by_name_singles = [
    {'name': 'Unknown', 'product': 'keyboard', 'amount':1},
    {'name': 'Yuki', 'product': 'manual', 'amount':1},
    {'name': 'anonymous', 'product': 'adopter', 'amount':2},
    {'name': 'Park', 'product': 'battery', 'amount':2},
    {'name': 'Smith', 'product': 'mouse', 'amount':1},
]


userTable_orderby_location_age_desc = [
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'},
    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
]

userTable_groupby_location_gender = [
    {'location': 'Korea', 'gender': 'M', 'count': 2, 'avg_age': 23.5},
    {'location': 'USA', 'gender': 'M', 'count': 2, 'avg_age': 18.0},
    {'location': 'China', 'gender': 'F', 'count': 1, 'avg_age': 28.0},
    {'location': 'China', 'gender': 'M', 'count': 1, 'avg_age': 15.0},
    {'location': 'Korea', 'gender': 'F', 'count': 1, 'avg_age': 12.0}
]


user_buy_table_inner_joined = [
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'cleaner', 'amount': 3},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'keyboard', 'amount': 1},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'monitor', 'amount': 1},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'mouse', 'amount': 3},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'hardcase', 'amount': 2},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'keycover', 'amount': 2},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'cleaner', 'amount': 5},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'mouse', 'amount': 1},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'product': 'battery', 'amount': 2},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'product': 'mouse', 'amount': 1},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'battery', 'amount': 7},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'cable', 'amount': 1},
]

user_buy_table_left_joined = [
    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA', 'product': '', 'amount': 0.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'cleaner', 'amount': 3.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'keyboard', 'amount': 1.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'monitor', 'amount': 1.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'mouse', 'amount': 3.0},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'hardcase', 'amount': 2.0},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'keycover', 'amount': 2.0},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'cleaner', 'amount': 5.0},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'mouse', 'amount': 1.0},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'product': 'battery', 'amount': 2.0},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'product': 'mouse', 'amount': 1.0},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'battery', 'amount': 7.0},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'cable', 'amount': 1.0},
]

user_buy_table_right_joined = [
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'cleaner', 'amount': 3.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'keyboard', 'amount': 1.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'monitor', 'amount': 1.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'mouse', 'amount': 3.0},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'hardcase', 'amount': 2.0},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'keycover', 'amount': 2.0},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'cleaner', 'amount': 5.0},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'mouse', 'amount': 1.0},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'product': 'battery', 'amount': 2.0},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'product': 'mouse', 'amount': 1.0},
    {'name': 'Unknown', 'gender': '', 'age': '', 'location': '', 'product': 'keyboard', 'amount': 1.0},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'battery', 'amount': 7.0},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'cable', 'amount': 1.0},
    {'name': 'Yuki', 'gender': '', 'age': '', 'location': '', 'product': 'manual', 'amount': 1.0},
    {'name': 'anonymous', 'gender': '', 'age': '', 'location': '', 'product': 'adopter', 'amount': 2.0},
]

user_buy_table_outer_joined = [
    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA', 'product': '', 'amount': 0.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'cleaner', 'amount': 3.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'keyboard', 'amount': 1.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'monitor', 'amount': 1.0},
    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea', 'product': 'mouse', 'amount': 3.0},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'hardcase', 'amount': 2.0},
    {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea', 'product': 'keycover', 'amount': 2.0},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'cleaner', 'amount': 5.0},
    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China', 'product': 'mouse', 'amount': 1.0},
    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea', 'product': 'battery', 'amount': 2.0},
    {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA', 'product': 'mouse', 'amount': 1.0},
    {'name': 'Unknown', 'gender': '', 'age': '', 'location': '', 'product': 'keyboard', 'amount': 1.0},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'battery', 'amount': 7.0},
    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China', 'product': 'cable', 'amount': 1.0},
    {'name': 'Yuki', 'gender': '', 'age': '', 'location': '', 'product': 'manual', 'amount': 1.0},
    {'name': 'anonymous', 'gender': '', 'age': '', 'location': '', 'product': 'adopter', 'amount': 2.0},
]


buyTable_values_count_by_product = {
    'adopter': 1,
    'battery': 2,
    'cable': 1,
    'cleaner': 2,
    'hardcase': 1,
    'keyboard': 2,
    'keycover': 1,
    'manual': 1,
    'monitor': 1,
    'mouse': 3
}







import inspect



class Undefined(object):
    """ Undefined objects always and reliably "do nothing." """

    def __init__(self, default=None):
        self.default = default
    def __call__(self, *args, **kwargs): return self
    def __repr__(self): return "undefined"
    def __bool__(self): return False
    def __eq__(self, __o: object) -> bool:
        return self.default == __o.default
    def __hash__(self) -> int:
        return hash(self.default)



def get_argcounts(callable):
    try:
        sig = inspect.signature(callable)
    except:
        return 1
    else:
        return len(sig.parameters)




from listorm import read_csv, select, groupby, asmap, fillmissed, read_excel
import re, math



def _split_amtunit(lst, columns, pfix_amt='amt', pfix_unit='unit'):
    pat = re.compile(r"(?P<amount>\d*\.?\d+)\s*(?P<unit>\w+)")
    def extract_amount(value):
        if g := pat.search(value):
            return g.group('amount')
    def extract_unit(value):
        if g:= pat.search(value):
            return g.group('unit')
        
    if isinstance(columns, str):
        columns = [columns]
    for column in columns:
        column_amt = f"{column}_{pfix_amt}"
        column_unit = f"{column}_{pfix_unit}"
        lst = lst.add_column(
            keymap={
                column_amt: lambda **row: extract_amount(row[column]),
                column_unit: lambda **row: extract_unit(row[column])
            }
        )
        lst = lst.set_number_type(formats={column_amt: 0.0})
#         lst = lst.add_column(**{column_amt: extract_amount, column_unit: extract_unit})
    return lst

if __name__ == '__main__':

    PRN_COLUMNS = [
        '발행처', '등록번호', '환자명', 'No.', '약품코드', '약품명', '투여량', '횟수', '일수', '총량', '용법', 'ADC', 'PRN', '퇴원', '처방의사', '입력일시',
        '투약번호', '차수', '접수일시'
    ]
    INJ_GROUPS = ['고가약', '고위험', '냉장약', '일반2', '마약', '향정약']

    lstdrug = read_excel('EUMC20210827약품정보.xlsx', fields_contains=['약품코드'], uniques='한글약품명' )
    lstprn = read_csv('prn_sample.txt', fields=PRN_COLUMNS, encoding='cp949',  delimiter='	')
    prn = lstprn.join(lstdrug, left_on='약품명', right_on='상용약품명').select(['발행처', '약품명', '기본투여단위', '함량', '규격', '환산단위', '투여량', '주사그룹번호(입)', '입력일시'])

    prn = _split_amtunit(prn, ['투여량', '함량', '규격'])

    prn = prn.add_column(수량=None)

    prn = prn.update(
        수량=lambda 투여량_amt, 함량_amt: 투여량_amt/함량_amt,
        where=lambda 함량_unit, 투여량_unit, 함량_amt: 함량_amt > 0 and  함량_unit == 투여량_unit
    )
    
    prn = prn.update(
        수량=lambda 투여량_amt, 규격_amt: 투여량_amt/규격_amt,
        where=lambda 수량, 규격_unit, 투여량_unit: 수량 is None and 규격_unit == 투여량_unit
    )

    prn = prn.update(
        수량=lambda 투여량_amt: 투여량_amt,
        where=lambda 환산단위, 투여량_unit, 수량: 수량 is None and 환산단위 == 투여량_unit
    )

    prn = prn.select(where=lambda 수량: 수량).update(수량=math.ceil)

    prn = prn.select(where=lambda **row: row['주사그룹번호(입)'] in INJ_GROUPS)

    grouped = prn.groupby(
        '약품명', '환산단위', '주사그룹번호(입)',
        수량=sum,
    ).orderby('주사그룹번호(입)', '약품명')

    grouby_word = prn.groupby(
        '주사그룹번호(입)', '약품명','발행처',
        수량=sum
    ).orderby('주사그룹번호(입)', '약품명')

    # grouped.print(1000)

    grouby_word.print(100)


import inspect



def tuplize(arg):
    if arg is None:
        return (arg)
    if isinstance(arg, (list, tuple)):
        return tuple(arg)
    return arg,


def str2float(value):
    try:
        ret = float(value)
    except:
        return value
    else:
        return ret

def round_try(value, round_to=2):
    try:
        if not isinstance(value, int):
            r = float(value)
        r = value
    except:
        return value
    else:
        return round(r, round_to)


def number_format(value, formats:object):
    if formats is None:
        return value
    formatter = type(formats)
    try:
        formatted = formatter(value)
    except Exception as e:
        asfloat = str2float(value)
        try:
            formatted = formatter(asfloat)
        except:
            formatted = formats
    return formatted


def get_argcounts(callable):
    try:
        sig = inspect.signature(callable)
    except:
        return 1
    else:
        return len(sig.parameters)



from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
	name='listorm',
	version='1.0.1',
	license='MIT',
	description='list orm methods and shortcuts for table type of dict-list',
	long_description=long_description,
	long_description_content_type="text/markdown",
	author = 'HS Moon',
	author_email = 'pbr112@naver.com',
	py_modules = ['listorm'],
	keywords=['listorm', 'list of dict', 'dict list', 'records', 'sql', 'orm'],
	url='https://github.com/zwolf21/listorm',
	packages=find_packages(exclude=['contrib', 'docs', 'tests']),
	install_requires=['openpyxl'],
	classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
	python_requires=">=3.6"
)



from ..utils import reduce_kwargs, reduce_args
from ..api import values_count, write_excel, write_csv, diff, set_number_format



class ShortCutMixin:

    @property
    def exists(self):
        return len(self) != 0

    @property
    def first(self):
        if self:
            return self[0]
    
    @property
    def count(self):
        return len(self)

    @property
    def last(self):
        if self:
            return self[-1]

    def max(self, column:str):
        '''get maximum of column's values

        :param column: column name where get max
        :return: max of column values


        .. doctest::
        
            >>> userTable = [
            ...     {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
            ...     {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
            ...     {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
            ...     {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
            ...     {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
            ...     {'name': 'Smith', 'gender': 'M', 'age': 17, 'location': 'USA'},
            ...     {'name': 'Lee', 'gender': 'F', 'age': 12, 'location': 'Korea'},
            ... ]

            >>> from listorm import Listorm
            >>> ls = Listorm(userTable)
            >>> ls.max('age')
            29
        '''
        values = filter(None, self.values(column))
        return max(values)

    def min(self, column:str):
        '''get minium of column's values

        :param column: column name where get min
        :return: min of column values


        .. doctest::

            >>> ls.min('age')
            12

        '''
        values = filter(None, self.values(column))
        return min(values)
    
    @reduce_kwargs('formats')
    def set_number_type(self, formats:dict):
        '''change number formats as default examples

        :param formats: number format examples
        :return: number formatted records


        .. doctest::

            >>> numbers = [
            ...     {
            ...       'flt': 0.5, 'string': '123', 'string_float': '123.5', 'int': 412, 'string_int': '5123', 'blabla': 'what?',  
            ...     }
            ... ]
            >>> lsnumber = Listorm(numbers)

            >>> # key name and example of types to change, if faild to change, example value will be default value
            >>> lsnumber.set_number_type(flt='', string=0.0, string_float=0, int='', string_int=0, blabla=0)
            [{'flt': '0.5', 'string': 123.0, 'string_float': 123, 'int': '412', 'string_int': 5123, 'blabla': 0}]

        '''
        records = set_number_format(self, formats=formats)
        return self.__class__(records)
        
    @reduce_args('columns')
    def values_count(self, columns:list):
        '''get values count by keys

        :param columns: column name for counting values
        :return: dict that has value occurance

        .. doctest::
        
            >>> lsb.values_count('product')
            {'battery': 2, 'keyboard': 2, 'cleaner': 2, 'monitor': 1, 'mouse': 3, 'hardcase': 1, 'keycover': 1, 'manual': 1, 'cable': 1, 'adopter': 1}

        '''
        return values_count(self, columns)
        
    def to_excel(self, filename=None):
        '''write to excel

        :param filename: output excel filename,  defaults to None
        :return: if filename is None, return file contents
        '''
        return write_excel(self, filename, fill_miss=False)


    def to_csv(self, filename=None, fields=None, encoding='utf-8', **csv_kwargs):
        '''write to csv

        :param filename: output csv filename, defaults to None
        :param fields: fields for retrieve, defaults to None
        :param encoding: encoding , defaults to 'utf-8'
        :return: if filename is None, return file contents
        '''
        return write_csv(self, filename, fields=fields, encoding=encoding, **csv_kwargs)

    def get_changes(self, other, pk=None):
        '''compare two records about added, deleted and updated on common columns

        :param other: another Listorm objects as records
        :param pk: unique keys for both of two records, defaults to None
        :return: namedtuple objects as Changes



        .. doctest::


            >>> before = [
            ...    {'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'},
            ...    {'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'},
            ...    {'name': 'ohmyboss', 'gender': 'F', 'age': 17, 'location': 'USA'},
            ...    {'name': 'Lyn', 'gender': 'F', 'age': 29, 'location': 'China'},
            ... ]
            >>> after = [
            ...    {'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'},
            ...    {'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'},
            ...    {'name': 'ohmyboss', 'gender': 'M', 'age': 17, 'location': 'Canada'},
            ...    {'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'},
            ... ]

            >>> ls_before = Listorm(before)
            >>> ls_after = Listorm(after)

            >>> changes = ls_before.get_changes(ls_after, pk='name')

            >>> changes.added
            [Added(pk='Xiaomi', rows={'name': 'Xiaomi', 'gender': 'M', 'age': 15, 'location': 'China'}), Added(pk='Park', rows={'name': 'Park', 'gender': 'M', 'age': 29, 'location': 'Korea'})]

            >>> changes.deleted
            [Deleted(pk='Hong', rows={'name': 'Hong', 'gender': 'M', 'age': 18, 'location': 'Korea'}), Deleted(pk='Charse', rows={'name': 'Charse', 'gender': 'M', 'age': 19, 'location': 'USA'})]

            >>> changes.updated
            [Updated(pk='ohmyboss', before={'name': 'ohmyboss', 'gender': 'F', 'age': 17, 'location': 'USA'}, after={'name': 'ohmyboss', 'gender': 'M', 'age': 17, 'location': 'Canada'}, where=['gender', 'location']), Updated(pk='Lyn', before={'name': 'Lyn', 'gender': 'F', 'age': 29, 'location': 'China'}, after={'name': 'Lyn', 'gender': 'F', 'age': 28, 'location': 'China'}, where=['age'])]

            >>> # usage of updated
            >>> for change in changes.updated:
            ...     print(change.pk)
            ...     for key in change.where:
            ...         print('where:', key)
            ...         print('  before->after:',change.before[key], '->', change.after[key])
            ...     print('------------------------------')
            ohmyboss
            where: gender
              before->after: F -> M
            where: location
              before->after: USA -> Canada
            ------------------------------
            Lyn
            where: age
              before->after: 29 -> 28
            ------------------------------

        '''
        if not pk and not all([self.uniques, other.uniques]):
            raise ValueError("Both of list has unique keys or specify pk for compare")
        elif not pk and self.uniques != other.uniques:
            raise ValueError("Both unique key fields({}, {}) must be same".format(self.uniques, other.uniques))
        elif not pk and not isinstance(other, self.__class__):
            raise ValueError("{} must be Listorm instance, not {}".format(other, type(other)))
        elif not all([self.exists, other.exists]):
            raise ValueError("Empty list not allowed:")
        
        diff_keys = pk or self.uniques or other.uniques
        return diff(self, other, diff_keys)